/* Groestl-opt.c     October 2008
 * ANSI C code optimised for 64-bit machines
 * Authors: Soeren S. Thomsen
 *          Krystian Matusiewicz
 *
 * This code is placed in the public domain
 */

#include "Groestl-opt.h"
#include "tables.h"

/* compute one new state column */
#define COLUMN(x,y,i,a,b,c,d,e,f,g,h,r,s)	\
  y[i] =					\
    T[(r)^EXT_BYTE(x[a],0)]^			\
    T[1*256+EXT_BYTE(x[b],1)]^			\
    T[2*256+EXT_BYTE(x[c],2)]^			\
    T[3*256+EXT_BYTE(x[d],3)]^			\
    T[4*256+EXT_BYTE(x[e],4)]^			\
    T[5*256+EXT_BYTE(x[f],5)]^			\
    T[6*256+EXT_BYTE(x[g],6)]^			\
    T[7*256+((s)^EXT_BYTE(x[h],7))]

/* compute a round in P or Q (short variants) */
#define RND512(x,y,r,s) do {			\
    COLUMN(x,y,7,7,0,1,2,3,4,5,6,0,0);		\
    COLUMN(x,y,6,6,7,0,1,2,3,4,5,0,0);		\
    COLUMN(x,y,5,5,6,7,0,1,2,3,4,0,0);		\
    COLUMN(x,y,4,4,5,6,7,0,1,2,3,0,0);		\
    COLUMN(x,y,3,3,4,5,6,7,0,1,2,0,0);		\
    COLUMN(x,y,2,2,3,4,5,6,7,0,1,0,0);		\
    COLUMN(x,y,1,1,2,3,4,5,6,7,0,0,s);		\
    COLUMN(x,y,0,0,1,2,3,4,5,6,7,r,0);		\
  } while (0)


/* compute a round in P or Q (long variants) */
#define RND1024(x,y,r,s) do {			\
    COLUMN(x,y,15,15, 0, 1, 2, 3, 4, 5,10,0,0);	\
    COLUMN(x,y,14,14,15, 0, 1, 2, 3, 4, 9,0,0);	\
    COLUMN(x,y,13,13,14,15, 0, 1, 2, 3, 8,0,0);	\
    COLUMN(x,y,12,12,13,14,15, 0, 1, 2, 7,0,0);	\
    COLUMN(x,y,11,11,12,13,14,15, 0, 1, 6,0,0);	\
    COLUMN(x,y,10,10,11,12,13,14,15, 0, 5,0,0);	\
    COLUMN(x,y, 9, 9,10,11,12,13,14,15, 4,0,0);	\
    COLUMN(x,y, 8, 8, 9,10,11,12,13,14, 3,0,0);	\
    COLUMN(x,y, 7, 7, 8, 9,10,11,12,13, 2,0,0);	\
    COLUMN(x,y, 6, 6, 7, 8, 9,10,11,12, 1,0,0);	\
    COLUMN(x,y, 5, 5, 6, 7, 8, 9,10,11, 0,0,s);	\
    COLUMN(x,y, 4, 4, 5, 6, 7, 8, 9,10,15,0,0);	\
    COLUMN(x,y, 3, 3, 4, 5, 6, 7, 8, 9,14,0,0);	\
    COLUMN(x,y, 2, 2, 3, 4, 5, 6, 7, 8,13,0,0);	\
    COLUMN(x,y, 1, 1, 2, 3, 4, 5, 6, 7,12,0,0);	\
    COLUMN(x,y, 0, 0, 1, 2, 3, 4, 5, 6,11,r,0);	\
} while (0)


/* the compression function (short variants) */
void F512(u64 *h, const u64 *m) {
  u64 y[COLS512];
  u64 z[COLS512];
  u64 outQ[COLS512];
  u64 inP[COLS512];
  int i;
  
  for (i = 0; i < COLS512; i++) {
    inP[i] = h[i] ^ m[i];
  }
	
  /* compute Q(m) */
  RND512(m,y,0,0^0xff);
  RND512(y,z,0,1^0xff);
  RND512(z,y,0,2^0xff);
  RND512(y,z,0,3^0xff);
  RND512(z,y,0,4^0xff);
  RND512(y,z,0,5^0xff);
  RND512(z,y,0,6^0xff);
  RND512(y,z,0,7^0xff);
  RND512(z,y,0,8^0xff);
  RND512(y,outQ,0,9^0xff);
	
  /* compute P(h+m) */
  RND512(inP,z,0,0);
  RND512(z,y,1,0);
  RND512(y,z,2,0);
  RND512(z,y,3,0);
  RND512(y,z,4,0);
  RND512(z,y,5,0);
  RND512(y,z,6,0);
  RND512(z,y,7,0);
  RND512(y,z,8,0);
  RND512(z,y,9,0);
	
  /* h' == h + Q(m) + P(h+m) */
  for (i = 0; i < COLS512; i++) {
    h[i] ^= outQ[i] ^ y[i];
  }
}/* F512 */


/* the compression function (long variants) */
void F1024(u64 *h, const u64 *m) {
  u64 y[COLS1024];
  u64 z[COLS1024];
  u64 outQ[COLS1024];
  u64 inP[COLS1024];
  int i;
  
  for (i = 0; i < COLS1024; i++) {
    inP[i] = h[i] ^ m[i];
  }

  /* compute Q(m) */
  RND1024(m,y,0,0xff);
  for (i = 1; i < ROUNDS1024-1; i += 2) {
    RND1024(y,z,0,i^0xff);
    RND1024(z,y,0,(i+1)^0xff);
  }
  RND1024(y,outQ,0,(ROUNDS1024-1)^0xff);
	
  /* compute P(h+m) */
  RND1024(inP,z,0,0);
  for (i = 1; i < ROUNDS1024-1; i += 2) {
    RND1024(z,y, i,0);
    RND1024(y,z, i+1,0);
  }
  RND1024(z,y,ROUNDS1024-1,0);
	
  /* h' == h + Q(m) + P(h+m) */
  for (i = 0; i < COLS1024; i++) {
    h[i] ^= outQ[i] ^ y[i];
  }
} /* F1024 */


/* digest up to msglen bytes of input (full blocks only) */
void Transform(hashState *ctx, 
	       const u8 *input, 
	       int msglen) {
  /* determine variant, SHORT or LONG, and select underlying
     compression function based on the variant */
  void (*F)(u64*,const u64*);
  switch ( ctx->v ) {
  case SHORT : F = &F512; break;
  case LONG  : 
  default    : F = &F1024; break;
  }

  /* increment block counter */
  ctx->block_counter += msglen/ctx->statesize;
  
  /* digest message, one block at a time */
  for (; msglen >= ctx->statesize; 
       msglen -= ctx->statesize, input += ctx->statesize) {
    F(ctx->chaining,(u64*)input);
  }
}

/* given state h, do h <- P(h)+h */
void OutputTransformation(hashState *ctx) {
  int j;
  u64 temp[COLS1024];
  u64 y[COLS1024];
  u64 z[COLS1024];

  /* determine variant */
  switch (ctx->v) {
  case SHORT :
    RND512(ctx->chaining,y,0,0);
    RND512(y,z,1,0);
    RND512(z,y,2,0);
    RND512(y,z,3,0);
    RND512(z,y,4,0);
    RND512(y,z,5,0);
    RND512(z,y,6,0);
    RND512(y,z,7,0);
    RND512(z,y,8,0);
    RND512(y,temp,9,0);
    for (j = 0; j < COLS512; j++) {
      ctx->chaining[j] ^= temp[j];
    }
    break;
  case LONG  :
    RND1024(ctx->chaining,y,0,0);
    for (j = 1; j < ROUNDS1024-1; j += 2) {
      RND1024(y,z,j,0);
      RND1024(z,y,j+1,0);
    }
    RND1024(y,temp,ROUNDS1024-1,0);
    for (j = 0; j < COLS1024; j++) {
      ctx->chaining[j] ^= temp[j];
    }
    break;
  }
}

/* initialise context */
HashReturn Init(hashState* ctx,
		int hashbitlen) {
  /* output size (in bits) must be a positive integer less than or
     equal to 512, and divisible by 8 */
  if (hashbitlen <= 0 || (hashbitlen%8) || hashbitlen > 512)
    return BAD_HASHLEN;

  /* set number of state columns and state size depending on
     variant */
  if (hashbitlen <= 256) {
    ctx->columns = COLS512;
    ctx->statesize = SIZE512;
    ctx->v = SHORT;
  }
  else {
    ctx->columns = COLS1024;
    ctx->statesize = SIZE1024;
    ctx->v = LONG;
  }

  /* allocate memory for state and data buffer */
  ctx->chaining = calloc(ctx->statesize,1);
  ctx->buffer = malloc(ctx->statesize);
  if (ctx->chaining == NULL || ctx->buffer == NULL)
    return FAIL;

  /* set initial value */
  ctx->chaining[ctx->columns-1] = U64BIG((u64)hashbitlen);

  /* set other variables */
  ctx->hashbitlen = hashbitlen;
  ctx->buf_ptr = 0;
  ctx->block_counter = 0;
  ctx->bits_in_last_byte = 0;

  return SUCCESS;
}

/* update state with databitlen bits of input */
HashReturn Update(hashState* ctx,
		  const BitSequence* input,
		  DataLength databitlen) {
  int index = 0;
  int msglen = (int)(databitlen/8);
  int rem = (int)(databitlen%8);

  /* non-integral number of message bytes can only be supplied in the
     last call to this function */
  if (ctx->bits_in_last_byte) return FAIL;

  /* if the buffer contains data that has not yet been digested, first
     add data to buffer until full */
  if (ctx->buf_ptr) {
    while (ctx->buf_ptr < ctx->statesize && index < msglen) {
      ctx->buffer[(int)ctx->buf_ptr++] = input[index++];
    }
    if (ctx->buf_ptr < ctx->statesize) {
      /* buffer still not full, return */
      if (rem) {
	ctx->bits_in_last_byte = rem;
	ctx->buffer[(int)ctx->buf_ptr++] = input[index];
      }
      return SUCCESS;
    }

    /* digest buffer */
    ctx->buf_ptr = 0;
    Transform(ctx, ctx->buffer, ctx->statesize);
  }

  /* digest bulk of message */
  Transform(ctx, input+index, msglen-index);
  index += ((msglen-index)/ctx->statesize)*ctx->statesize;

  /* store remaining data in buffer */
  while (index < msglen) {
    ctx->buffer[(int)ctx->buf_ptr++] = input[index++];
  }

  /* if non-integral number of bytes have been supplied, store
     remaining bits in last byte, together with information about
     number of bits */
  if (rem) {
    ctx->bits_in_last_byte = rem;
    ctx->buffer[(int)ctx->buf_ptr++] = input[index];
  }
  return SUCCESS;
}

#define BILB ctx->bits_in_last_byte

/* finalise: process remaining data (including padding), perform
   output transformation, and write hash result to 'output' */
HashReturn Final(hashState* ctx,
		 BitSequence* output) {
  int i, j = 0, hashbytelen = ctx->hashbitlen/8;
  u8 *s = (BitSequence*)ctx->chaining;

  /* pad with '1'-bit and first few '0'-bits */
  if (BILB) {
    ctx->buffer[(int)ctx->buf_ptr-1] &= ((1<<BILB)-1)<<(8-BILB);
    ctx->buffer[(int)ctx->buf_ptr-1] ^= 0x1<<(7-BILB);
    BILB = 0;
  }
  else ctx->buffer[(int)ctx->buf_ptr++] = 0x80;

  /* pad with '0'-bits */
  if (ctx->buf_ptr > ctx->statesize-LENGTHFIELDLEN) {
    /* padding requires two blocks */
    while (ctx->buf_ptr < ctx->statesize) {
      ctx->buffer[(int)ctx->buf_ptr++] = 0;
    }
    /* digest first padding block */
    Transform(ctx, ctx->buffer, ctx->statesize);
    ctx->buf_ptr = 0;
  }
  while (ctx->buf_ptr < ctx->statesize-LENGTHFIELDLEN) {
    ctx->buffer[(int)ctx->buf_ptr++] = 0;
  }

  /* length padding */
  ctx->block_counter++;
  ctx->buf_ptr = ctx->statesize;
  while (ctx->buf_ptr > ctx->statesize-LENGTHFIELDLEN) {
    ctx->buffer[(int)--ctx->buf_ptr] = (u8)ctx->block_counter;
    ctx->block_counter >>= 8;
  }

  /* digest final padding block */
  Transform(ctx, ctx->buffer, ctx->statesize);
  /* perform output transformation */
  OutputTransformation(ctx);

  /* store hash result in output */
  for (i = ctx->statesize-hashbytelen; i < ctx->statesize; i++,j++) {
    output[j] = s[i];
  }

  /* zeroise relevant variables and deallocate memory */
  for (i = 0; i < ctx->columns; i++) {
    ctx->chaining[i] = 0;
  }
  for (i = 0; i < ctx->statesize; i++) {
    ctx->buffer[i] = 0;
  }
  free(ctx->chaining);
  free(ctx->buffer);

  return SUCCESS;
}

/* hash bit sequence */
HashReturn Hash(int hashbitlen,
		const BitSequence* data, 
		DataLength databitlen,
		BitSequence* hashval) {
  HashReturn ret;
  hashState context;

  /* initialise */
  if ((ret = Init(&context, hashbitlen)) != SUCCESS)
    return ret;

  /* process message */
  if ((ret = Update(&context, data, databitlen)) != SUCCESS)
    return ret;

  /* finalise */
  ret = Final(&context, hashval);

  return ret;
}

void PrintHash(const BitSequence* hash,
	       int hashbitlen) {
  int i;
  for (i = 0; i < hashbitlen/8; i++) {
    printf("%02x", hash[i]);
  }
  printf("\n");
}
